%Code associated with section V.B of https://hal.inria.fr/hal-01708695.
%
%This code runs with GraSP 1.1 (https://gforge.inria.fr/projects/grasp).
%
% Authors:
%  - Benjamin Girault <benjamin.girault@usc.edu>

% Copyright Benjamin Girault, University of Sourthern California, Los
% Angeles, California, USA (2018)
% 
% benjamin.girault@usc.edu
% 
% This software is governed by the CeCILL license under French law and
% abiding by the rules of distribution of free software.  You can  use, 
% modify and/ or redistribute the software under the terms of the CeCILL
% license as circulated by CEA, CNRS and INRIA at the following URL
% "http://www.cecill.info". 
% 
% As a counterpart to the access to the source code and  rights to copy,
% modify and redistribute granted by the license, users are provided only
% with a limited warranty  and the software's author,  the holder of the
% economic rights,  and the successive licensors  have only  limited
% liability. 
% 
% In this respect, the user's attention is drawn to the risks associated
% with loading,  using,  modifying and/or developing or reproducing the
% software by the user in light of its specific status of free software,
% that may mean  that it is complicated to manipulate,  and  that  also
% therefore means  that it is reserved for developers  and  experienced
% professionals having in-depth computer knowledge. Users are therefore
% encouraged to load and test the software's suitability as regards their
% requirements in conditions enabling the security of their systems and/or 
% data to be ensured and,  more generally, to use and operate it in the 
% same conditions as regards security. 
% 
% The fact that you are presently reading this means that you have had
% knowledge of the CeCILL license and that you accept its terms.

%% General Parameters
N = 1000;                   % Number of vertices
Nmin = 50;                  % Min Number of vertices (for a graph model with a random number of vertices)
Nmax = 200;                 % Max Number of vertices (for a graph model with a random number of vertices)

gaussian_sigma = 0.3;       % Gaussian weight parameter
KNN = 10;                   % Number of neighbors selected in KNN
dmax = 1;                   % For the thresholded selection scheme
gamma = 1;                  % Weight normalization by the degree

Nb_graph_realizations = 5;  % Number of realization of the graph model

signal_freqs = 0.5:0.5:10;  % Continuous signal frequencies to compute statistics on
% signal_freqs = 0.5:0.5:10;
% signal_freqs = 1:2:10;

signal_phase_shifts = (0:0.01:0.99) * 2 * pi;   % Continuous signal phase shifts to compute statistics on
% signal_phase_shifts = 0:0.1:(2 * pi);

%% Output folder
dir = 'MatlabData/';
if ~isdir(dir)
    mkdir(dir);
end

%% Subaxis Parameters
Margin  = 0.004;
Spacing = 0;
Padding = [0.01 0.01 0.02 0.02];

%% Helper function
get_data = @(smooth) reshape(smooth, [], size(smooth, 3));

%% Signal Model Definition

signal_model = @(x, y, freq, phase) cos(2 * pi * freq * x / 10 + phase);
model_name = '';

% signal_model = @(x, y, freq, phase) exp(1i * 2 * pi * freq * x / 10 + 1i * phase);
% model_name = 'exp_';

signal_model_graph = @(g, freq, phase) signal_model(g.layout(:, 1), g.layout(:, 2), freq, phase);

% random_phases = unifrnd(0, 2 * pi, Nb_signal_realizations, 1);



%% Samples Distribution (for a non uniform distribution of vertices)

% minimum prob
e = 0.2;

ramp_slope = @(start, finish, level_start, level_end) (level_start - level_end) / (start - finish);
ramp_bias  = @(start, finish, level_start, level_end) level_start - start * ramp_slope(start, finish, level_start, level_end);
ramp = @(x, start, finish, level_start, level_end) ramp_slope(start, finish, level_start, level_end) * x + ramp_bias(start, finish, level_start, level_end);

ramp_up_down = @(x, start, middle, finish, level_high, level_low)...
    (x >= start & x <= middle) .* ramp(x, start, middle, level_low, level_high) +...
    (x > middle & x <= finish) .* ramp(x, middle, finish, level_high, level_low);

ramp1_start = 1;
ramp1_middle = 2;
ramp1_end = 3;

ramp2_start = 7;
ramp2_middle = 8;
ramp2_end = 9;

p_start = @(x) e * double(x < ramp1_start);
p_ramp1 = @(x) ramp_up_down(x, ramp1_start, ramp1_middle, ramp1_end, 1, e);
p_middle = @(x) e * double(x > ramp1_end & x < ramp2_start);
p_ramp2 = @(x) ramp_up_down(x, ramp2_start, ramp2_middle, ramp2_end, 1, e);
p_end = @(x) e * double(x > ramp2_end);
p = @(x) p_start(x) + p_ramp1(x) + p_middle(x) + p_ramp2(x) + p_end(x);
p_name = 'double_ramp';

pdf = @(x) p(x) / (10 * e + (1 - e) * (ramp1_end - ramp1_start + ramp2_end - ramp2_start) / 2);

cdf = @(x) pdf(0) * x...
         + (x >= ramp1_middle)                  .* (pdf(ramp1_middle) - pdf(0))  * (ramp1_middle - ramp1_start) / 2 ...
         + (x >= ramp1_end)                     .* (pdf(ramp1_middle) - pdf(0))  * (ramp1_end - ramp1_middle) / 2 ...
         + (x >= ramp2_middle)                  .* (pdf(ramp1_middle) - pdf(0))  * (ramp2_middle - ramp2_start) / 2 ...
         + (x >= ramp2_end)                     .* (pdf(ramp1_middle) - pdf(0))  * (ramp2_end - ramp2_middle) / 2 ...
         + (x > ramp1_start & x < ramp1_middle) .* (pdf(x) - pdf(0))            .* (x - ramp1_start) / 2 ...
         + (x > ramp1_middle & x < ramp1_end)   .* (pdf(ramp1_middle) - pdf(x)) .* (x - ramp1_middle) / 2 ...
         + (x > ramp1_middle & x < ramp1_end)   .* (pdf(x) - pdf(0))            .* (x - ramp1_middle) ...
         + (x > ramp2_start & x < ramp2_middle) .* (pdf(x) - pdf(0))            .* (x - ramp2_start) / 2 ...
         + (x > ramp2_middle & x < ramp2_end)   .* (pdf(ramp2_middle) - pdf(x)) .* (x - ramp2_middle) / 2 ...
         + (x > ramp2_middle & x < ramp2_end)   .* (pdf(x) - pdf(0))            .* (x - ramp2_middle);

grasp_open_figure_name('Sampling Density');
x = (0:0.01:10)';
plot(x, [p(x) pdf(x) cdf(x)])
legend({'Density', 'PDF', 'CDF'});

csvwrite([dir 'sampling_distribution.csv'], [x / 10 p(x) pdf(x) cdf(x)]);

%% Plot of a continuous signals

tmp = [kron((0:300)', ones(301, 1)) kron(ones(301, 1), (0:300)')];

grasp_open_figure_name('Example Continuous Signals');

grasp_subaxis(2, 2, 1, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model(tmp(:, 1) / 30, tmp(:, 2) / 30, 1, 0);
imshow(reshape(x, 301, 301), [-1 1])
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=1, \phi=0', 'FontSize', 30);

grasp_subaxis(2, 2, 2, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model(tmp(:, 1) / 30, tmp(:, 2) / 30, 1, pi / 4);
imshow(reshape(x, 301, 301), [-1 1])
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=1, \phi=\pi/4', 'FontSize', 30);

grasp_subaxis(2, 2, 3, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model(tmp(:, 1) / 30, tmp(:, 2) / 30, 3, 0);
imshow(reshape(x, 301, 301), [-1 1])
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=3, \phi=0', 'FontSize', 30);

grasp_subaxis(2, 2, 4, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model(tmp(:, 1) / 30, tmp(:, 2) / 30, 3, pi / 4);
imshow(reshape(x, 301, 301), [-1 1])
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=3, \phi=\pi/4', 'FontSize', 30);


grasp_open_figure_name('Example Sampled Signals (unif)');

g = grasp_plane_rnd(300);

grasp_subaxis(2, 2, 1, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 1, 0);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=1, \phi=0', 'FontSize', 30);

grasp_subaxis(2, 2, 2, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 1, pi / 4);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=1, \phi=\pi/4', 'FontSize', 30);

grasp_subaxis(2, 2, 3, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 3, 0);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=3, \phi=0', 'FontSize', 30);

grasp_subaxis(2, 2, 4, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 3, pi / 4);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=3, \phi=\pi/4', 'FontSize', 30);


grasp_open_figure_name('Example Sampled Signals (non unif)');

g = grasp_struct;
g.layout = [];
while size(g.layout, 1) < 300
    new_layout = unifrnd(0, 10, 5 * (300 - size(g.layout, 1)), 2);
    new_mask = p(new_layout(:, 1)) >= unifrnd(0, 1, 5 * (300 - size(g.layout, 1)), 1);
    nb_new_samples = min(numel(new_mask), 300 - size(g.layout, 1));
    new_layout = new_layout(new_mask(1:nb_new_samples), :);
    g.layout = [g.layout ; new_layout];
end
g.A = zeros(300);

grasp_subaxis(2, 2, 1, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 1, 0);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=1, \phi=0', 'FontSize', 30);

grasp_subaxis(2, 2, 2, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 1, pi / 4);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=1, \phi=\pi/4', 'FontSize', 30);

grasp_subaxis(2, 2, 3, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 3, 0);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=3, \phi=0', 'FontSize', 30);

grasp_subaxis(2, 2, 4, 'M', Margin, 'S', Spacing, 'P', Padding, 'PT', 0.08);
x = signal_model_graph(g, 3, pi / 4);
grasp_show_graph(gca, g, 'node_values', x, 'show_edges', false);
grasp_set_blue_red_colormap(gca);
colorbar;
title('\nu=3, \phi=\pi/4', 'FontSize', 30);

%% Mean Data, for various setups

% mean_data = cell(0);

%% Sampling Realizations, with GFTs and Power Spectrum of the signal above

clear degree
degree     (Nb_graph_realizations, N) = 0;

clear l2_norm deg_norm voro_norm
l2_norm         (Nb_graph_realizations, numel(signal_phase_shifts), numel(signal_freqs)) = 0;
deg_norm        (Nb_graph_realizations, numel(signal_phase_shifts), numel(signal_freqs)) = 0;
voro_norm       (Nb_graph_realizations, numel(signal_phase_shifts), numel(signal_freqs)) = 0;

for cur_graph_realization = 1:Nb_graph_realizations
    %% Generate a graph from sampling the Euclidean plane
%     graph_name = sprintf('unif_%d', N);
%     cur_graph = grasp_plane_rnd(N);
%     fprintf('Uniform Distribution, fixed number of vertices (%d)\n', N);
    
%     graph_name = sprintf('unif_%d-%d', Nmin, Nmax);
%     cur_graph = grasp_plane_rnd(Nmin - 1 + randi(Nmax - Nmin));
%     fprintf('Uniform Distribution, variable number of vertices (%d-%d)\n', Nmin, Nmax);
    
    graph_name = sprintf('non_unif_%s_%d', p_name, N);
    cur_graph = grasp_struct;
    cur_graph.layout = [];
    while size(cur_graph.layout, 1) < N
        new_layout = unifrnd(0, 10, 5 * (N - size(cur_graph.layout, 1)), 2);
        new_mask = p(new_layout(:, 1)) >= unifrnd(0, 1, 5 * (N - size(cur_graph.layout, 1)), 1);
        nb_new_samples = min(numel(new_mask), N - size(cur_graph.layout, 1));
        new_layout = new_layout(new_mask(1:nb_new_samples), :);
        cur_graph.layout = [cur_graph.layout ; new_layout];
    end
    fprintf('Non-Uniform Distribution (%s), fixed number of vertices (%d)\n', p_name, N);

    %% Compute the base adjacency matrix
    cur_graph.distances = grasp_distances_layout(cur_graph);
    cur_graph.A = grasp_adjacency_gaussian(cur_graph, gaussian_sigma);
    
    %% Graph sparsification
%     sparse_name = '';
    
%     sparse_name = sprintf('_KNN_%d', KNN);
%     cur_graph.A = grasp_adjacency_knn(cur_graph, KNN);
%     fprintf('KNN (K=%d)\n', KNN);

    sparse_name = sprintf('_Thresh_%.2f', dmax);
    cur_graph.A = grasp_adjacency_thresh(cur_graph, exp(-(dmax) ^ 2 / (2 * gaussian_sigma^2)));
    fprintf('Thresholded Weights\n');
    
    %% Weight normalization    
%     weight_norm = '';

    weight_norm = '_wnorm';
    cur_graph = grasp_adjacency_degreenorm(cur_graph, gamma);
    fprintf('Weight Normalization (gamma=%f)\n', gamma);
    
%     weight_norm = '_unweighted';
%     cur_graph.A = double(cur_graph.A > 0);

    %% GFT
    cur_D    = grasp_degrees(cur_graph);
    cur_voro = diag(grasp_voronoi_areas(cur_graph));
        
    degree(cur_graph_realization, :)   = diag(cur_D);
    
    %% Test on the stationary continuous signal
    for freq_id = 1:numel(signal_freqs)
        cur_phase_shited_signals                     = signal_model_graph(cur_graph, signal_freqs(freq_id), signal_phase_shifts);
        l2_norm(cur_graph_realization, :, freq_id)   = diag(cur_phase_shited_signals' * cur_phase_shited_signals);
        deg_norm(cur_graph_realization, :, freq_id)  = diag(cur_phase_shited_signals' * cur_D * cur_phase_shited_signals);
        voro_norm(cur_graph_realization, :, freq_id) = diag(cur_phase_shited_signals' * cur_voro * cur_phase_shited_signals);
    end
end

cur_data_id = 1;
if exist('mean_data', 'var')
    cur_data_id = numel(mean_data) + 1;
end

mean_data{cur_data_id}.type      = [model_name graph_name sparse_name weight_norm];

mean_data{cur_data_id}.mean_degnorm   = mean(get_data(deg_norm));
mean_data{cur_data_id}.median_degnorm = median(get_data(deg_norm));
mean_data{cur_data_id}.std_degnorm    = std(get_data(deg_norm));

mean_data{cur_data_id}.mean_degree     = mean(degree(:));
mean_data{cur_data_id}.std_mean_degree = std(mean(degree));

mean_data{cur_data_id}.mean_l2norm    = reshape(mean(l2_norm), size(l2_norm, 2), size(l2_norm, 3));
mean_data{cur_data_id}.mean_deg_norm  = reshape(mean(deg_norm), size(deg_norm, 2), size(deg_norm, 3));
mean_data{cur_data_id}.mean_voro_norm = reshape(mean(voro_norm), size(voro_norm, 2), size(voro_norm, 3));

mean_data{cur_data_id}.var_l2norm    = reshape(var(l2_norm), size(l2_norm, 2), size(l2_norm, 3));
mean_data{cur_data_id}.var_deg_norm  = reshape(var(deg_norm), size(deg_norm, 2), size(deg_norm, 3));
mean_data{cur_data_id}.var_voro_norm = reshape(var(voro_norm), size(voro_norm, 2), size(voro_norm, 3));

%% Save an example graph

grasp_exportcsv(cur_graph, [dir mean_data{cur_data_id}.type '_nodes.csv'], [dir mean_data{cur_data_id}.type '_edges.csv']);
grasp_exportcsv_signal(cur_graph, diag(grasp_degrees(cur_graph)), [dir mean_data{cur_data_id}.type '_degrees.csv']);
grasp_exportcsv_signal(cur_graph, grasp_voronoi_areas(cur_graph), [dir mean_data{cur_data_id}.type '_voronoi.csv']);

%% Save the Data on Energy Depending on Frequency and Phase Shift

% mu^(Q)_nu
mean_data{cur_data_id}.rescale_mean_l2norm    = mean(mean_data{cur_data_id}.mean_l2norm);
mean_data{cur_data_id}.rescale_mean_deg_norm  = mean(mean_data{cur_data_id}.mean_deg_norm);
mean_data{cur_data_id}.rescale_mean_voro_norm = mean(mean_data{cur_data_id}.mean_voro_norm);

csvwrite([dir mean_data{cur_data_id}.type '_norms_mean_over_phases.csv'],...
         [signal_freqs' ...
          mean_data{cur_data_id}.rescale_mean_l2norm'    / mean(mean_data{cur_data_id}.rescale_mean_l2norm)...
          mean_data{cur_data_id}.rescale_mean_deg_norm'  / mean(mean_data{cur_data_id}.rescale_mean_deg_norm)...
          mean_data{cur_data_id}.rescale_mean_voro_norm' / mean(mean_data{cur_data_id}.rescale_mean_voro_norm)]);

% E^(Q)_nu(phi) / mu^(Q)_nu
mean_data{cur_data_id}.norm_mean_l2norm    = mean_data{cur_data_id}.mean_l2norm * diag(mean_data{cur_data_id}.rescale_mean_l2norm .^ (-1));
mean_data{cur_data_id}.norm_mean_deg_norm  = mean_data{cur_data_id}.mean_deg_norm * diag(mean_data{cur_data_id}.rescale_mean_deg_norm .^ (-1));
mean_data{cur_data_id}.norm_mean_voro_norm = mean_data{cur_data_id}.mean_voro_norm * diag(mean_data{cur_data_id}.rescale_mean_voro_norm .^ (-1));

csvwrite([dir mean_data{cur_data_id}.type '_l2norms_phase_normalized.csv'],   [0 signal_freqs ; signal_phase_shifts' mean_data{cur_data_id}.norm_mean_l2norm]);
csvwrite([dir mean_data{cur_data_id}.type '_degnorms_phase_normalized.csv'],  [0 signal_freqs ; signal_phase_shifts' mean_data{cur_data_id}.norm_mean_deg_norm]);
csvwrite([dir mean_data{cur_data_id}.type '_voronorms_phase_normalized.csv'], [0 signal_freqs ; signal_phase_shifts' mean_data{cur_data_id}.norm_mean_voro_norm]);

csvwrite([dir mean_data{cur_data_id}.type '_l2norms_phase_max_abs_dev.csv'],   [signal_freqs' max(abs(1 - mean_data{cur_data_id}.norm_mean_l2norm))']);
csvwrite([dir mean_data{cur_data_id}.type '_degnorms_phase_max_abs_dev.csv'],  [signal_freqs' max(abs(1 - mean_data{cur_data_id}.norm_mean_deg_norm))']);
csvwrite([dir mean_data{cur_data_id}.type '_voronorms_phase_max_abs_dev.csv'], [signal_freqs' max(abs(1 - mean_data{cur_data_id}.norm_mean_voro_norm))']);

mean_data{cur_data_id}.cv_l2norm    = sqrt(mean_data{cur_data_id}.var_l2norm) ./ mean_data{cur_data_id}.mean_l2norm;
mean_data{cur_data_id}.cv_deg_norm  = sqrt(mean_data{cur_data_id}.var_deg_norm) ./ mean_data{cur_data_id}.mean_deg_norm;
mean_data{cur_data_id}.cv_voro_norm = sqrt(mean_data{cur_data_id}.var_voro_norm) ./ mean_data{cur_data_id}.mean_voro_norm;

csvwrite([dir mean_data{cur_data_id}.type '_l2norms_max_cv.csv'], [signal_freqs' max(mean_data{cur_data_id}.cv_l2norm)']);
csvwrite([dir mean_data{cur_data_id}.type '_degnorms_max_cv.csv'], [signal_freqs' max(mean_data{cur_data_id}.cv_deg_norm)']);
csvwrite([dir mean_data{cur_data_id}.type '_voronorms_max_cv.csv'], [signal_freqs' max(mean_data{cur_data_id}.cv_voro_norm)']);


%% Plots

grasp_open_figure_name(sprintf('Mean Energy (%s)', mean_data{cur_data_id}.type));
plot(signal_freqs, [mean_data{cur_data_id}.rescale_mean_l2norm'    / mean(mean_data{cur_data_id}.rescale_mean_l2norm) ...
             mean_data{cur_data_id}.rescale_mean_deg_norm'  / mean(mean_data{cur_data_id}.rescale_mean_deg_norm) ...
             mean_data{cur_data_id}.rescale_mean_voro_norm' / mean(mean_data{cur_data_id}.rescale_mean_voro_norm)],...
             'LineWidth', 3)
legend({'l2', 'deg', 'voro'})

grasp_open_figure_name(sprintf('CV & Maxdev (%s)', mean_data{cur_data_id}.type));

grasp_subaxis(2, 1, 1, 'M', Margin, 'S', 0.03, 'P', Padding, 'PT', 0.08, 'MH', 0.06, 'MB', 0.03);
plot(signal_freqs, max(abs(1 - mean_data{cur_data_id}.norm_mean_l2norm))', 'LineWidth', 3)
legend({'l2'}, 'FontSize', 30)
plot(signal_freqs, [max(abs(1 - mean_data{cur_data_id}.norm_mean_l2norm))' max(abs(1 - mean_data{cur_data_id}.norm_mean_deg_norm))' max(abs(1 - mean_data{cur_data_id}.norm_mean_voro_norm))'], 'LineWidth', 3)
legend({'l2', 'deg', 'voro'}, 'FontSize', 30)
title('MaxDev', 'FontSize', 30);

grasp_subaxis(2, 1, 2, 'M', Margin, 'S', 0.03, 'P', Padding, 'PT', 0.08, 'MH', 0.06, 'MB', 0.03);
plot(signal_freqs, max(mean_data{cur_data_id}.cv_l2norm)', 'LineWidth', 3)
legend({'l2'}, 'FontSize', 30, 'Location', 'SouthEast')
plot(signal_freqs, [max(mean_data{cur_data_id}.cv_l2norm)' max(mean_data{cur_data_id}.cv_deg_norm)' max(mean_data{cur_data_id}.cv_voro_norm)'], 'LineWidth', 3)
legend({'l2', 'deg', 'voro'}, 'FontSize', 30, 'Location', 'SouthEast')
title('CV', 'FontSize', 30);

